/*
*	Filename:		MeshExport.cpp
*	Author:			Lars Vidar Magnusson
*	Copyright (C)	LaViMa Development 2006
*/

#include <MeshExport.h>

cMeshExport::cMeshExport() {

}

cMeshExport::~cMeshExport() {

}

bool cMeshExport::isAlreadyExported(MDagPath &meshPath) {
	// Find parent string
	MString fullPathName = meshPath.fullPathName();
	int endIndex = fullPathName.rindex('|');
	MString parentString = fullPathName.substring(0, endIndex);

	// Check if it's already exported
	bool found = false;
	std::vector<MString>::iterator iter;
	for(iter=mAlreadyExported.begin(); iter!=mAlreadyExported.end(); ++iter) {
		if(*iter == parentString) {
			found = true;
		}
	}
	if(!found) {
		mAlreadyExported.push_back(parentString);
	}
	return found;
}

MStatus cMeshExport::extractVertices(MFnMesh &fnMesh, tMesh *mesh) {
	MStatus status;

	MPointArray vertices;
	status = fnMesh.getPoints(vertices, MSpace::kWorld);
	mesh->vertices.reserve(vertices.length());
	for(unsigned int i=0; i<vertices.length(); ++i) {
		mesh->vertices.push_back(vertices[i]);
	}
	return status;
}

MStatus cMeshExport::extractNormals(MFnMesh &fnMesh, tMesh *mesh) {
	MStatus status;

	MFloatVectorArray normals;
	status = fnMesh.getNormals(normals, MSpace::kWorld);
	mesh->normals.reserve(normals.length());
	for(unsigned int i=0; i<normals.length(); ++i) {
		mesh->normals.push_back(normals[i]);
	}
	return status;
}

MStatus cMeshExport::extractUVs(MFnMesh &fnMesh, tMesh *mesh) {
	MStatus status;

	MFloatArray us;
	MFloatArray vs;

	status = fnMesh.getUVs(us, vs);
	mesh->uvs.reserve(us.length());
	
	for(unsigned int j=0; j<us.length(); ++j) {
		tUV currentUV;
		currentUV.u = us[j];
		currentUV.v = vs[j];
		mesh->uvs.push_back(currentUV);
	}
	return status;
}

MStatus cMeshExport::extractWeights(MDagPath meshPath, tMesh *mesh) {
	MStatus status;

	MFnMesh fnMesh(meshPath);

	MItDependencyNodes depNodeIt(MFn::kSkinClusterFilter);
	for( ; !depNodeIt.isDone(); depNodeIt.next()) {
		
		MObject depNodeObject = depNodeIt.item();
    MFnSkinCluster fnSkinCluster(depNodeObject, &status);

		// Check if the mesh is part of the skins output
		unsigned int shapeIndex = fnSkinCluster.indexForOutputShape(meshPath.node(), &status);
    if(!status) 
			return status;
		
		MObject mayaInputObject = fnSkinCluster.inputShapeAtIndex(shapeIndex, &status);

		MDagPathArray jointNames;
		fnSkinCluster.influenceObjects(jointNames, &status);

		MItGeometry geometryIt(mayaInputObject);
		unsigned int vertexIndex = 0;
		for ( ; !geometryIt.isDone(); geometryIt.next(), ++vertexIndex) {
			
			MObject mayaComponent = geometryIt.component();
			MFloatArray mayaWeightArray;
			unsigned int numWeights;

			// Not sure how the the mayaWeightArray is sorted, but I'm assuming it's depth-first with this code
      fnSkinCluster.getWeights(meshPath, mayaComponent, mayaWeightArray, numWeights);
			for(unsigned int jointIndex=0; jointIndex<mayaWeightArray.length(); ++jointIndex) {
				if(mayaWeightArray[jointIndex] != 0) {
					tInfluence newInfluence;
					newInfluence.jointIndex = jointIndex;
					newInfluence.jointName = jointNames[jointIndex].partialPathName(&status);
					newInfluence.vertexIndex = vertexIndex;
					newInfluence.weight = mayaWeightArray[jointIndex];
					mesh->influences.push_back(newInfluence);
				}
			}
		}
	}
	return status;
}

MStatus cMeshExport::extractTriangles(MFnMesh &fnMesh, tMesh *mesh, MIntArray &attributes) {
	MStatus status = MStatus::kSuccess;

	unsigned int i = 0;
	MItMeshPolygon polygonIter(fnMesh.object(), &status);
	for( ; !polygonIter.isDone(); polygonIter.next(), ++i) {
		MIntArray verticesArray;
		status = polygonIter.getVertices(verticesArray);

		mesh->shaderIndices.push_back(attributes[i]);

		// Make sure the mesh is triangulated
		if(verticesArray.length() == 3) {
			for(unsigned int i=0; i<verticesArray.length(); ++i) {
        mesh->vertexIndices.push_back(verticesArray[i]);
				
				int normalIndex = polygonIter.normalIndex(i, &status);
				mesh->normalIndices.push_back(normalIndex);
				
				int uvIndex;
				status = polygonIter.getUVIndex(verticesArray[i], uvIndex);
				mesh->uvIndices.push_back(uvIndex);
			}
		}
	}
	return status;
}

MStatus cMeshExport::extractMaterials(MObjectArray &shaders) {
	MStatus status = MStatus::kSuccess;

	// Loop through all the shaders
	for(unsigned int index=0; index<shaders.length(); ++index) {
		MFnDependencyNode depNodeShader(shaders[index], &status);
		
		MPlug surfaceShaderPlug = depNodeShader.findPlug("surfaceShader", &status);
		if(surfaceShaderPlug.isNull()) continue;

		MPlugArray shaderPlugSources;
		surfaceShaderPlug.connectedTo(shaderPlugSources, true, false);

		tShader *newShader = new tShader();

		for(unsigned int source=0; source<shaderPlugSources.length(); ++source) {
			MPlug currentSource = shaderPlugSources[source];
			MObject materialNode = currentSource.node(&status);

			// Only support lambert shaders for now
			if(materialNode.hasFn(MFn::kLambert)) {
				tMaterial *newMaterial = new tMaterial();
				
				MFnDependencyNode depNodeMaterial(materialNode, &status);

				MPlug colorR = depNodeMaterial.findPlug("ambientColorR");
				MPlug colorG = depNodeMaterial.findPlug("ambientColorG");
				MPlug colorB = depNodeMaterial.findPlug("ambientColorB");
				MPlug colorA;

				status = colorR.getValue(newMaterial->ambient.r);
				status = colorG.getValue(newMaterial->ambient.g);
				status = colorB.getValue(newMaterial->ambient.b);
				newMaterial->ambient.a = 1.0f;

				bool useOutColor = false;
				MPlug color = depNodeMaterial.findPlug("color");
				if(color.isNull()) {
					color = depNodeMaterial.findPlug("outColor");
					useOutColor= true;
				}

				bool textureFound = false;
				MPlugArray connectedObjects;
				if(!color.isNull() && color.connectedTo(connectedObjects, true, false)) {
					MPlug texture= connectedObjects[0];

					MObject objectNode = texture.node();

					if(objectNode.hasFn( MFn::kFileTexture)) {
						MFnDependencyNode textureNode(objectNode);
						MPlug file = textureNode.findPlug("fileTextureName");
						file.getValue(newMaterial->diffuseMap);
						
						textureFound = true;
					}
				}

				newShader->materials.push_back(newMaterial);
			}
		}
		mShaders.push_back(newShader);
	}

	return status;
}

MStatus cMeshExport::extractMesh(MDagPath &meshPath) {
	MStatus status = MStatus::kSuccess;
	tMesh *newMesh = new tMesh;

	// This test is potential dangerous. If the skinned mesh is found later than the 
	// original it will be skipped, but it seems to work.
	if(!isAlreadyExported(meshPath)) {
	  MFnMesh fnMesh(meshPath);

		unsigned int instanceNumber = meshPath.instanceNumber(&status);
		MObjectArray shaders;
		MIntArray attributes;
		status = fnMesh.getConnectedShaders(instanceNumber, shaders, attributes);

		MString name = meshPath.partialPathName(&status);
		newMesh->name = name.asChar();

		status = extractVertices(fnMesh, newMesh);
		status = extractNormals(fnMesh, newMesh);
		status = extractUVs(fnMesh, newMesh);
		status = extractTriangles(fnMesh, newMesh, attributes);
		status = extractWeights(meshPath, newMesh);

		status = extractMaterials(shaders);
		
		mMeshes.push_back(newMesh);
	}

	return status;
}

MStatus cMeshExport::extractMeshes() {
	MStatus status = MStatus::kSuccess;

	MItDag dagIter(MItDag::kDepthFirst, MFn::kMesh, &status);
	for( ; !dagIter.isDone(); dagIter.next()) {
		MDagPath currentPath; 
		status = dagIter.getPath(currentPath);
		extractMesh(currentPath);		
	}	
	return status;
}

void cMeshExport::writeMeshes(cTextFileWriter &writer) {
	std::vector<tMesh *>::iterator iter;
	for(iter=mMeshes.begin(); iter!=mMeshes.end(); ++iter) {
		writer.addTab();
		writer.add("mesh \"%s\"\n", (*iter)->name.c_str());

		writer.addTab();
		writer.add("vertices %i\n", (*iter)->vertices.size()); 
		for(unsigned int i=0; i<(*iter)->vertices.size(); ++i) {
			writer.add("%i %f %f %f\n", i, (*iter)->vertices[i].x, (*iter)->vertices[i].y, (*iter)->vertices[i].z);
		}
    writer.removeTab();
		writer.add("\n");

		writer.addTab();
    writer.add("normals %i\n", (*iter)->normals.size()); 
		for(unsigned int i=0; i<(*iter)->normals.size(); ++i) {
			writer.add("%i %f %f %f\n", i, (*iter)->normals[i][0], (*iter)->normals[i][1], (*iter)->normals[i][2]);
		}
		writer.removeTab();
		writer.add("\n");

		writer.addTab();
		writer.add("uvs %i\n", (*iter)->uvs.size());
		for(unsigned int i=0; i<(*iter)->uvs.size(); ++i) {
			writer.add("%i %f %f\n", i, (*iter)->uvs[i].u, (*iter)->uvs[i].v);
		}
		writer.removeTab();
		writer.add("\n");

		writer.addTab();
		writer.add("triangles %i\n", (*iter)->vertexIndices.size()/3);
		for(unsigned int i=0; i<(*iter)->vertexIndices.size(); i+=3) {
			int v = i;
			int n = i;
			int t = i;
			writer.add("%i %i %i %i %i %i %i %i %i %i %i\n", i/3, (*iter)->shaderIndices[i/3], 
				(*iter)->vertexIndices[i], (*iter)->vertexIndices[i+1], (*iter)->vertexIndices[i+2],
				(*iter)->normalIndices[i], (*iter)->normalIndices[i+1], (*iter)->normalIndices[i+2],
				(*iter)->uvIndices[i], (*iter)->uvIndices[i+1], (*iter)->uvIndices[i+2]);
		}
		writer.removeTab();
		writer.add("\n");

		writer.addTab();
		writer.add("weights %i\n", (*iter)->influences.size());
		for(unsigned int i=0; i<(*iter)->influences.size(); ++i) {
			writer.add("%i %i %i \"%s\" %f\n", i, (*iter)->influences[i].vertexIndex, (*iter)->influences[i].jointIndex, (*iter)->influences[i].jointName.asChar(), (*iter)->influences[i].weight);
		}
		writer.removeTab();
		writer.add("\n");

		writer.removeTab();
		writer.add("\n");
	}
}

void cMeshExport::writeMaterials(cTextFileWriter &writer) {
	std::vector<tShader *>::iterator shadeIter;
	for(shadeIter=mShaders.begin(); shadeIter!=mShaders.end(); ++shadeIter) {

		std::vector<tMaterial *>::iterator matIter;
		for(matIter=(*shadeIter)->materials.begin(); matIter!=(*shadeIter)->materials.end(); ++matIter) {
      writer.add("ambient %f %f %f %f\n", (*matIter)->ambient.r, (*matIter)->ambient.g, (*matIter)->ambient.b, (*matIter)->ambient.a);
			writer.add("diffusemap %s\n", (*matIter)->diffuseMap.asChar());
		}
	}
}

void cMeshExport::cleanUp() {
	std::vector<tMesh *>::iterator meshIter;
	for(meshIter=mMeshes.begin(); meshIter!=mMeshes.end(); ++meshIter) {
		delete *meshIter;
	}
	mMeshes.clear();

	std::vector<tShader *>::iterator shadeIter;
	for(shadeIter=mShaders.begin(); shadeIter!=mShaders.end(); ++shadeIter) {
		
		std::vector<tMaterial *>::iterator matIter;
		for(matIter=(*shadeIter)->materials.begin(); matIter!=(*shadeIter)->materials.end(); ++matIter) {
			delete *matIter;
		}
		(*shadeIter)->materials.clear();
		delete *shadeIter;
	}
	mShaders.clear();
	mAlreadyExported.clear();
}